#include "model.h"

#include <QSize>

ModelWindow::ModelWindow(QWindow *parent)
    : QWindow(parent)
{
    setSurfaceType(QWindow::OpenGLSurface);

	// Se intenta abrir la cámara en OpenCV
	int cols = 640;
	int rows = 480;
	int deviceID = 4;  // 0 = open default camera
	// int apiID = cv::CAP_ANY;
	int apiID = cv::CAP_V4L;

		 // Initiating camera
	cap = new VideoCapture();
	cap->open(deviceID + apiID);  // open the  camera

	if (!cap->isOpened()) {  // check if we succeeded
		printf("Error while opening camera device cap(1)\n");
		exit(0);
	}
	else{
		printf("Ok camera\n");
	}
	cap->set(CAP_PROP_FRAME_WIDTH, cols);
	cap->set(CAP_PROP_FRAME_HEIGHT, rows);

	timer = new QTimer( this );
	connect ( timer, SIGNAL(timeout()), this, SLOT(timerDone()) );
}

/**
ModelWindow::~ModelWindow()
{
    delete m_device;
}
**/


void ModelWindow::initialize( )
{
    m_program = new QOpenGLShaderProgram(this);

    m_program->addShaderFromSourceFile(QOpenGLShader::Vertex, "vertexShaderSource.vertexshader" );
    m_program->addShaderFromSourceFile(QOpenGLShader::Fragment, "texture.fragmentshader" );  
    m_program->link();

    m_posAttr = m_program->attributeLocation("posAttr");
    Q_ASSERT(m_posAttr != -1);

    m_matrixUniform = m_program->uniformLocation("matrix");
    Q_ASSERT(m_matrixUniform != -1);

    m_texCoord = m_program->attributeLocation("texCoord");
    Q_ASSERT(m_texCoord != -1);

	m_program->bind();
	m_program->setUniformValue("texture", 0);

	inicializa_matrices();

	// Para la imagen de la cámara:
	texture = new QOpenGLTexture(QOpenGLTexture::Target2D);  //3553 de1 
	// Allocates server-side storage for this texture object taking into account, 
	// the format, dimensions, mipmap levels, array layers and cubemap faces.
	texture->setFormat(QOpenGLTexture::RGBA8_UNorm ); //QOpenGLTexture::BGR
	texture->setSize(640,480,1);
	texture->setMipLevels(10);
	//  texture->setLayers(1); Texture target does not support array layers
	texture->allocateStorage(QOpenGLTexture::BGR, QOpenGLTexture::UInt8);

    if ( objImg.load( "textura.png", "PNG" ) == false ) {
        fprintf( stderr, "ERROR: no se pudo cargar la imagen de textura\n" );
    }
    objTexture = new QOpenGLTexture( objImg.mirrored() );
}


//! [5]
void ModelWindow::render( )
{
	// glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// Se lee una imagen de la cámara
	 *(cap) >> frame;  // get a new frame from camera
	flip(frame, frame, 0);
	// cv::cvtColor(frame, grayFrame, COLOR_BGR2GRAY);
    //  void    setData(QOpenGLTexture::PixelFormat sourceFormat, QOpenGLTexture::PixelType sourceType, const void *data, const QOpenGLPixelTransferOptions *const options = nullptr)
	texture->setData(QOpenGLTexture::BGR , QOpenGLTexture::UInt8, frame.data);

    const qreal retinaScale = devicePixelRatio();
    glViewport(0, 0, width() * retinaScale, height() * retinaScale);

    
	// Se dibuja la imagen de fondo
    m_program->bind();

	QMatrix4x4 matrix, M;
	matrix.ortho(0.0, 640.0, 0.0, 480.0, 0, 41);

	m_program->setUniformValue(m_matrixUniform, matrix);

	static const GLfloat vertices_back[] = {
	      0.0f,     0.0f, -40.0,
	    640.0f,     0.0f, -40.0,
	    640.0f,   480.0f, -40.0,
	      0.0f,     0.0f, -40.0,
	    640.0f,   480.0f, -40.0,
	      0.0f,   480.0f, -40.0,
	};
 	static const GLfloat uvs_back[] = {
	    0.0f,  0.0f,
	    1.0f,  0.0f,
	    1.0f,  1.0f,
	    0.0f,  0.0f,
	    1.0f,  1.0f,
	    0.0f,  1.0f
	};
	
    // m_program->setUniformValue(m_colorUniform, QColor( 210, 210, 210, 0) );
    glEnableVertexAttribArray(m_posAttr);
    glVertexAttribPointer(m_posAttr, 3, GL_FLOAT, GL_FALSE, 0, vertices_back);

	glEnableVertexAttribArray(m_texCoord);
    glVertexAttribPointer( m_texCoord, 2, GL_FLOAT, GL_FALSE, 0, uvs_back );

	texture->bind();
    glDrawArrays(GL_TRIANGLES, 0, 6);

    glDisableVertexAttribArray(m_posAttr);
    glDisableVertexAttribArray(m_texCoord);
	
	/** El objeto **/

    M = matrix;
	// M.translate( 320.0, 240.0, 0.0 );
	M.translate( 120.0, 120.0, 0.0 );
	M.scale( 8.0, 8.0, 1.0 );
	M.lookAt( QVector3D( 5.0, 5.0, 10.0), QVector3D( 0.0, 0.0, 0.0), QVector3D( 0.0, 1.0, 0.0) );
	// M.scale( 10.0, 10.0, 10.0 );
	// M.translate( 320.0, 240.0, 0.0 );
    
    m_program->setUniformValue(m_matrixUniform, M);

    static const GLfloat vertices[] = {
         -10.0f,  -10.0f, 0.0,
          10.0f,  -10.0f, 0.0,
          10.0f,   10.0f, 0.0,
         -10.0f,  -10.0f, 0.0,
          10.0f,   10.0f, 0.0,
         -10.0f,   10.0f, 0.0,
    };

    // m_program->setUniformValue(m_colorUniform, QColor( 255, 255, 0, 0) );

    glVertexAttribPointer(m_posAttr, 3, GL_FLOAT, GL_FALSE, 0, vertices);
    glEnableVertexAttribArray(m_posAttr);

	glEnableVertexAttribArray(m_texCoord);
    glVertexAttribPointer( m_texCoord, 2, GL_FLOAT, GL_FALSE, 0, uvs_back );

	objTexture->bind();
	// texture->bind();
    glDrawArrays(GL_TRIANGLES, 0, 6);

    glDisableVertexAttribArray(m_posAttr);
    glDisableVertexAttribArray(m_texCoord);

    m_program->release();
}
//! [5]

void ModelWindow::inicializa_matrices()
{
    MK[0]  = f;    MK[1]  = 0.0; MK[2]  = -255.0; MK[3]  = 0.0;  
    MK[4]  = 0.0;  MK[5]  = f;   MK[6]  = -255.0; MK[7]  = 0.0;  
    MK[8]  = 0.0;  MK[9]  = 0.0; MK[10] = 101.0;  MK[11] = 100.0;  
    MK[12] = 0.0;  MK[13] = 0.0; MK[14] = -1.0;   MK[15] = 0.0;  

    K = QMatrix4x4 (MK);
    
    MRT[0 ] = 0.9736;  MRT[1 ] = 0.2282;  MRT[2 ] = -0.0002;  MRT[3 ] =  0.0382 ;    
    MRT[4 ] = -0.0895; MRT[5 ] = 0.3809;  MRT[6 ] =  0.9203;  MRT[7 ] = -0.025; 
    MRT[8 ] = 0.21;    MRT[9 ] = -0.896;  MRT[10] = 0.3913;   MRT[11] = -83.6709;
    MRT[12 ] = 0.0;    MRT[13 ] = 0.0;    MRT[14] = 0.0;      MRT[15] = 1.0;
    RT = QMatrix4x4 (MRT);

	/**
	const float * p = RT.constData();
	for( int i=0; i<16; i++ )
		printf("%g ", p[i] );
	printf("\n" );
	**/
}

void ModelWindow::keyPressEvent( QKeyEvent *e )
{
	switch ( e->key() ) {
		case Qt::Key_Escape:
		case Qt::Key_Q:
			QCoreApplication::quit();
			break;
		case Qt::Key_I:
			timer->start(33);
			break;
		case Qt::Key_C: // Captura lo que está viendo la cámara en una imagen
			capture( );
			break;

	}
}

bool ModelWindow::event(QEvent *event)
{   
    switch (event->type()) {
    case QEvent::UpdateRequest:
        renderNow();
        return true;
    default:
        return QWindow::event(event);
    }
}

void ModelWindow::exposeEvent(QExposeEvent *event)
{   
    Q_UNUSED(event);
    
    if (isExposed())
        renderNow();
}

void ModelWindow::renderNow()
{
    if (!isExposed())
        return;

    bool needsInitialize = false;

    if (!m_context) {
        m_context = new QOpenGLContext(this);
        m_context->setFormat(requestedFormat());
        m_context->create();

        needsInitialize = true;
    }

    m_context->makeCurrent(this);

    if (needsInitialize) {
        initializeOpenGLFunctions();
		initialize( );
    }
    
    render();

    m_context->swapBuffers(this);
}

void ModelWindow::capture()
{
	flip(frame, frame, 0);
	imwrite("imagen.png", frame);
}

void ModelWindow::timerDone( )
{
	renderNow( );
}

